import json
import time

import requests

from .utils import get_domain
from .utils.errors import (AuthenticationException,
                           RefreshTokenExpirationException,
                           catch_requests_error, ensure_response_ok)

ONE_THOUSAND_YEARS_IN_SECONDS = 1000 * 365 * 24 * 3600


class Authentication:
    """
    Class to authenticate in the ELG using out-of-band authentication.

    Args:
        base_url (str): url to authenticate.
        token_url (str): url to request tokens.
        client (str): name of the Keycloak client.
        redirect_uri (str): redirect uri.

    """

    base_url = "{}/auth/realms/ELG/protocol/openid-connect/auth"
    token_url = "{}/auth/realms/ELG/protocol/openid-connect/token"
    client = "elg-oob"
    redirect_uri = "urn:ietf:wg:oauth:2.0:oob"

    def __init__(self, domain: str):
        """
        Init an Authentication object.

        Args:
            domain (str): ELG domain you want to use. "live" to use the public ELG, "dev" to use the development ELG
                and another value to use a local ELG.
        """
        self.access_token = None
        self.expires_in = None
        self.expires_time = time.gmtime(0)
        self.refresh_expires_in = None
        self.refresh_expires_time = time.gmtime(0)
        self.refresh_token = None
        self.token_type = None
        self.id_token = None
        self.not_before_policy = None
        self.session_state = None
        self.scope = None
        self.domain = get_domain(domain=domain)

    def __str__(self):
        return f"Authentication(valid_until: {time.strftime('%Y-%m-%dT%H:%M:%SZ', self.expires_time)})"

    def __repr__(self):
        obj = self.__dict__.copy()
        obj["access_token"] = "*****"
        obj["refresh_token"] = "*****"
        obj["id_token"] = "*****"
        return str(obj)

    @classmethod
    def init(cls, scope: str = "openid", domain: str = "live"):
        """
        Class method to init an Authentication object and authenticate to the ELG.

        Args:
            scope (str, optional): scope to use when requesting tokens. Can be set to "openid" or "offline_access"
                to get offline tokens. Defaults to "openid".
            domain (str, optional): ELG domain you want to use. "live" to use the public ELG, "dev" to use the
                development ELG and another value to use a local ELG. Defaults to "live".

        Returns:
            elg.Authentication: Authentication object with Keycloak authentication tokens.
        """
        authentication = cls(domain=domain)
        authentication.create(scope)
        return authentication

    @classmethod
    def from_json(cls, filename: str):
        """
        Class method to init an Authentication object from a json file.

        Args:
            filename (str): name of the json file.

        Returns:
            elg.Authentication: Authentication object with Keycloak authentication tokens.
        """
        with open(filename, "r") as tokens_file:
            tokens = json.load(tokens_file)
        authentication = cls(tokens["domain"])
        authentication.access_token = tokens["access_token"]
        authentication.expires_in = tokens["expires_in"]
        authentication.expires_time = time.struct_time(tokens["expires_time"])
        authentication.refresh_expires_in = tokens["refresh_expires_in"]
        authentication.refresh_expires_time = time.struct_time(tokens["refresh_expires_time"])
        authentication.refresh_token = tokens["refresh_token"]
        authentication.token_type = tokens["token_type"]
        authentication.id_token = tokens["id_token"]
        authentication.not_before_policy = tokens["not_before_policy"]
        authentication.session_state = tokens["session_state"]
        authentication.scope = tokens["scope"]
        return authentication

    @classmethod
    def create_authentication_url(cls, scope: str = "openid", domain: str = "live") -> str:
        """
        Class method to create the Keycloak authentication url.

        Args:
            scope (str, optional): scope to use when requesting tokens. Can be set to "openid" or "offline_access"
                to get offline tokens. Defaults to "openid".
            domain (str, optional): ELG domain you want to use. "live" to use the public ELG, "dev" to use the
                development ELG and another value to use a local ELG. Defaults to "live".

        Returns:
            str: Keycloak authentication url.
        """
        domain = get_domain(domain=domain)
        queries = [
            ("client_id", cls.client),
            ("redirect_uri", cls.redirect_uri),
            ("response_type", "code"),
            ("scope", scope),
        ]
        auth_url = (
            cls.base_url.format(domain)
            + ("?" if len(queries) >= 1 else "")
            + "&".join([f"{query}={str(value)}" for (query, value) in queries])
        )
        return auth_url

    @classmethod
    def from_success_code(cls, code: str, domain: str = "live"):
        """
        Class method to init an Authentication object from a success code.

        Args:
            code (str): success code obtained after authentication.
            domain (str, optional): ELG domain you want to use. "live" to use the public ELG, "dev" to use the
                development ELG and another value to use a local ELG. Defaults to "live".

        Returns:
            elg.Authentication: Authentication object with Keycloak authentication tokens.
        """
        authentication = cls(domain=domain)
        data = {
            "grant_type": "authorization_code",
            "code": code,
            "redirect_uri": cls.redirect_uri,
            "client_id": cls.client,
        }
        authentication._requesting_oauth_token(data)
        return authentication

    def create(self, scope: str = "openid"):
        """
        Method to create Keycloak authentication tokens.

        Args:
            scope (str, optional): scope to use when requesting tokens. Can be set to "openid" or "offline_access"
                to get offline tokens. Defaults to "openid".
        """
        queries = [
            ("client_id", self.client),
            ("redirect_uri", self.redirect_uri),
            ("response_type", "code"),
            ("scope", scope),
        ]
        auth_url = (
            self.base_url.format(self.domain)
            + ("?" if len(queries) >= 1 else "")
            + "&".join([f"{query}={str(value)}" for (query, value) in queries])
        )

        print("Please go to this URL in your browser: {}\n".format(auth_url))
        auth_code = input('Paste the "success code": ')
        data = {
            "grant_type": "authorization_code",
            "code": auth_code,
            "redirect_uri": self.redirect_uri,
            "client_id": self.client,
        }
        self._requesting_oauth_token(data=data)

    def refresh(self):
        """
        Method to refresh to access_token using the refresh_token.

        Raises:
            AuthenticationException: There is no refresh_token to refresh the access_token.
        """
        if self.refresh_token is None:
            raise AuthenticationException()
        data = {"grant_type": "refresh_token", "client_id": self.client, "refresh_token": self.refresh_token}
        self._requesting_oauth_token(data=data)

    def refresh_if_needed(self):
        """
        Method that call the refresh method only if needed, i.e. the access_token is expired.

        Raises:
            RefreshTokenExpirationException: The refresh_token is expired.
        """
        # refresh if the token will expire in less than 60 seconds
        # idea from: https://stackoverflow.com/questions/30826726/how-to-identify-if-the-oauth-token-has-expired
        if time.gmtime(time.time() + 60) > self.expires_time:
            if time.gmtime() > self.refresh_expires_time:
                raise RefreshTokenExpirationException()
            if time.gmtime(time.time() + 3600) > self.refresh_expires_time:
                token_expire_time = time.mktime(time.gmtime()) - time.mktime(self.refresh_expires_time)
                print(f"Warning: The refresh token will expire in {token_expire_time} seconds!")
            self.refresh()

    @catch_requests_error
    def _requesting_oauth_token(self, data):
        """
        Method to obtain the Keycloak authentication tokens from data.

        Args:
            data (dict): data used to ask the tokens to Keycloak.
        """
        response = requests.post(self.token_url.format(self.domain), data=data)
        ensure_response_ok(response=response)
        tokens = response.json()
        self.access_token = tokens["access_token"]
        self.expires_in = tokens["expires_in"]
        self.expires_time = time.gmtime(time.time() + self.expires_in)
        self.refresh_expires_in = tokens["refresh_expires_in"]
        self.refresh_expires_time = time.gmtime(
            time.time() + (self.refresh_expires_in if self.refresh_expires_in != 0 else ONE_THOUSAND_YEARS_IN_SECONDS)
        )
        self.refresh_token = tokens["refresh_token"]
        self.token_type = tokens["token_type"]
        self.id_token = tokens.get(
            "id_token"
        )  # use get to not raise an error if there is no id_token (for offline_access)
        self.not_before_policy = tokens["not-before-policy"]
        self.session_state = tokens["session_state"]
        self.scope = tokens["scope"]

    def to_json(self, filename: str):
        """
        Save the Keycloak authentication tokens to a json file.

        Args:
            filename (str): name of the json file.
        """
        with open(filename, "w") as tokens_file:
            json.dump(self.__dict__, tokens_file, indent="\t")
