import hashlib
import io
import pickle
import time
from os import makedirs
from os.path import expanduser, isfile, splitext
from pathlib import Path
from time import sleep
from typing import Callable, List, Union

import requests

from .authentication import Authentication
from .entity import Entity
from .utils import (MIME, get_argument_from_json, get_content_from_response,
                    get_domain, get_information, get_metadatarecord,
                    map_metadatarecord_to_result)
from .utils.errors import (DomainException, NotServiceException,
                           RefreshTokenExpirationException,
                           catch_requests_error, ensure_response_ok)


def need_authentication():
    """
    Decorator for Service methods to refresh to authentication tokens before calling the method
    """

    def decorator(func):
        def wrapper(self, *args, **kwargs):
            self.authentication.refresh_if_needed()
            result = func(self, *args, **kwargs)
            return result

        return wrapper

    return decorator


class Service(Entity):
    """
    Class to use ELG service
    """

    def __init__(
        self,
        id: int,
        resource_name: str,
        resource_short_name: List[str],
        resource_type: str,
        entity_type: str,
        description: str,
        keywords: List[str],
        detail: str,
        licences: List[str],
        languages: List[str],
        country_of_registration: List[str],
        creation_date: str,
        last_date_updated: str,
        functional_service: bool,
        functions: List[str],
        intended_applications: List[str],
        views: int,
        downloads: int,
        size: int,
        service_execution_count: int,
        status: str,
        under_construction: bool,
        record: dict,
        auth_object: Authentication,
        auth_file: str,
        scope: str,
        domain: str,
        use_cache: bool,
        cache_dir: str,
    ):
        """
        Init a Service object with the service information
        """
        super().__init__(
            id=id,
            resource_name=resource_name,
            resource_short_name=resource_short_name,
            resource_type=resource_type,
            entity_type=entity_type,
            description=description,
            keywords=keywords,
            detail=detail,
            licences=licences,
            languages=languages,
            country_of_registration=country_of_registration,
            creation_date=creation_date,
            last_date_updated=last_date_updated,
            functional_service=functional_service,
            functions=functions,
            intended_applications=intended_applications,
            views=views,
            downloads=downloads,
            size=size,
            service_execution_count=service_execution_count,
            status=status,
            under_construction=under_construction,
            domain=domain,
            record=record,
        )
        if self.resource_type != "ToolService" and self.resource_type != "Tool/Service":
            raise NotServiceException(self.id)
        self._authenticate(
            auth_object=auth_object, auth_file=auth_file, scope=scope, use_cache=use_cache, cache_dir=cache_dir
        )
        self._get_elg_execution_location()
        self._get_resource_info()

    @classmethod
    def from_id(
        cls,
        id: int,
        auth_object: Authentication = None,
        auth_file: str = None,
        scope: str = None,
        domain: str = None,
        use_cache: bool = True,
        cache_dir: str = "~/.cache/elg",
    ):
        """
        Class method to init a Service class from its id. You can provide authentication information through the
        auth_object or the auth_file attributes. If not authentication information is provided, the Authentication
        object will be initialized.

        Args:
            id (int): id of the service.
            auth_object (elg.Authentication, optional): elg.Authentication object to use. Defaults to None.
            auth_file (str, optional): json file that contains the authentication tokens. Defaults to None.
            scope (str, optional): scope to use when requesting tokens. Can be set to "openid" or "offline_access"
                to get offline tokens. Defaults to "openid".
            domain (str, optional): ELG domain you want to use. "live" to use the public ELG, "dev" to use the
                development ELG and another value to use a local ELG. Defaults to "live".
            use_cache (bool, optional): True if you want to use cached files. Defaults to True.
            cache_dir (str, optional): path to the cache_dir. Set it to None to not store any cached files. Defaults to "~/.cache/elg".

        Returns:
            elg.Service: Service object with authentication information.
        """
        # Use auth_object/file domain if auth_object/file given
        if auth_object is not None:
            assert isinstance(
                auth_object, Authentication
            ), f"auth_object ({type(auth_object)}) must be an Authentication object."
            auth_domain = auth_object.domain
            if domain is not None:
                print(
                    "Warning: domain argument saved in the authentication object will overwrite the "
                    "domain argument you put."
                )
            domain = auth_domain
        elif isinstance(auth_file, str):
            auth_domain = get_argument_from_json(auth_file, "domain")
            if domain is not None:
                print(
                    "Warning: domain argument saved in the authentication filename will overwrite the "
                    "domain argument you put."
                )
            domain = auth_domain
        # Use "live" domain by default
        domain = get_domain(domain=domain if domain is not None else "live")
        record = get_metadatarecord(id=id, domain=domain, use_cache=use_cache, cache_dir=cache_dir)
        result = map_metadatarecord_to_result(id=id, record=record)
        result["auth_object"] = auth_object
        result["auth_file"] = auth_file
        result["scope"] = scope
        result["domain"] = domain
        result["use_cache"] = use_cache
        result["cache_dir"] = cache_dir
        return cls(**result)

    @classmethod
    def from_entity(
        cls,
        entity: Entity,
        auth_object: str = None,
        auth_file: str = None,
        scope: str = None,
        use_cache: bool = True,
        cache_dir="~/.cache/elg",
    ):
        """
        Class method to init a Service class from an Entity object. You can provide authentication information through the
        auth_object or the auth_file attributes. If not authentication information is provided, the Authentication object will be initialized.

        Args:
            entity (elg.Entity): Entity object to init as a Service.
            auth_object (elg.Authentication, optional): elg.Authentication object to use. Defaults to None.
            auth_file (str, optional): json file that contains the authentication tokens. Defaults to None.
            scope (str, optional): scope to use when requesting tokens. Can be set to "openid" or "offline_access"
                to get offline tokens. Defaults to "openid".
            domain (str, optional): ELG domain you want to use. "live" to use the public ELG, "dev" to use the
                development ELG and another value to use a local ELG. Defaults to "live".
            use_cache (bool, optional): True if you want to use cached files. Defaults to True.
            cache_dir (str, optional): path to the cache_dir. Set it to None to not store any cached files. Defaults to "~/.cache/elg".

        Returns:
            elg.Service: Service object with authentication information.
        """
        if entity.record is None:
            entity.record = get_metadatarecord(
                id=entity.id, domain=entity.domain, use_cache=use_cache, cache_dir=cache_dir
            )
        parameters = entity.__dict__
        parameters["auth_object"] = auth_object
        parameters["auth_file"] = auth_file
        parameters["scope"] = scope
        parameters["use_cache"] = use_cache
        parameters["cache_dir"] = cache_dir
        return cls(**parameters)

    def _authenticate(self, auth_object: Authentication, auth_file: str, scope: str, use_cache: str, cache_dir: str):
        """
        Method to authenticate.

        Args:
            auth_object (elg.Authentication, optional): elg.Authentication object to use.
            auth_file (str, optional): json file that contains the authentication tokens.
            scope (str, optional): scope to use when requesting tokens. Can be set to "openid" or "offline_access"
                to get offline tokens.
            domain (str, optional): ELG domain you want to use. "live" to use the public ELG, "dev" to use the
                development ELG and another value to use a local ELG.
            use_cache (bool, optional): True if you want to use cached files.
            cache_dir (str, optional): path to the cache_dir. Set it to None to not store any cached files.

        Raises:
            DomainException: the domain of the authentication is different from the service domain.
        """

        # Use the authentification object if given
        if auth_object is not None:
            print(f"Using authentication object")
            assert isinstance(
                auth_object, Authentication
            ), f"auth_object ({type(auth_object)}) must be an Authentication object."
            if auth_file is not None:
                print("Warning: auth_file argument will not be used as an auth_object has been passed.")
            if scope is not None:
                print(
                    "Warning: scope argument saved in the authentication object will overwrite the "
                    "scope argument you put."
                )
            self.authentication = auth_object
            if self.domain != self.authentication.domain:
                raise DomainException(self.domain, self.authentication.domain)
            if not self._is_auth_object_valid(self.authentication):
                raise RefreshTokenExpirationException()

        # Use the authentication file if given
        elif isinstance(auth_file, str):
            print(f"Using authentication file: {auth_file}")
            if scope is not None:
                print(
                    "Warning: scope argument saved in the authentication filename will overwrite the "
                    "scope argument you put."
                )
            self.authentication = Authentication.from_json(filename=auth_file)
            if self.domain != self.authentication.domain:
                raise DomainException(self.domain, self.authentication.domain)
            if not self._is_auth_object_valid(self.authentication):
                raise RefreshTokenExpirationException()

        # Use a cached authentification object or create a new one
        else:
            scope = scope if scope is not None else "openid"  # default scope

            if cache_dir is not None and use_cache is True:
                cache_path = expanduser(cache_dir)
                cache_path = Path(cache_path) / Path("tokens")
                cache_path.mkdir(parents=True, exist_ok=True)
                authentication_hash = hashlib.sha256(
                    bytes(f"authentication - {scope} - {self.domain}", "utf-8")
                ).hexdigest()
                cache_filename = cache_path / Path(authentication_hash)
                if cache_filename.is_file():
                    with open(cache_filename, "rb") as cache_file:
                        self.authentication = pickle.load(cache_file)

            if (not hasattr(self, "authentication")) or (not self._is_auth_object_valid(self.authentication)):
                self.authentication = Authentication.init(scope=scope, domain=self.domain)

        if cache_dir is not None:
            cache_path = expanduser(cache_dir)
            cache_path = Path(cache_path) / Path("tokens")
            cache_path.mkdir(parents=True, exist_ok=True)
            authentication_hash = hashlib.sha256(
                bytes(
                    f"authentication - {'openid' if 'openid' in self.authentication.scope else 'offline_access'} - {self.authentication.domain}",
                    "utf-8",
                )
            ).hexdigest()
            cache_filename = cache_path / Path(authentication_hash)
            with open(cache_filename, "wb") as cache_file:
                pickle.dump(self.authentication, cache_file)

    def _is_auth_file_valid(self, auth_file: str, scope: str = None) -> bool:
        """
        Method to check is an authentication file contains valid authentication information.

        Args:
            auth_file (str): json file that contains the authentication tokens.
            scope (str, optional): scope to use when requesting tokens. Can be set to "openid" or "offline_access"
                to get offline tokens.

        Returns:
            bool: True is the authentication file contains valid information, else False.
        """
        refresh_expires_time, auth_domain, auth_scope = get_argument_from_json(
            filename=auth_file, args=["refresh_expires_time", "domain", "scope"]
        )
        # verify if tokens are valid
        if time.gmtime() > time.struct_time(refresh_expires_time):
            return False
        # verify if the domain of cached tokens corresponds to the desire domain
        if self.domain != auth_domain:
            return False
        # verify if the scope of cached tokens corresponds to the desire scope
        if scope is not None and scope not in auth_scope:
            return False
        return True

    def _is_auth_object_valid(self, auth_object: Authentication, scope: str = None) -> bool:
        """
        Method to check is an authentication object contains valid authentication information.

        Args:
            auth_object (Authentication): authentication object.
            scope (str, optional): scope to use when requesting tokens. Can be set to "openid" or "offline_access"
                to get offline tokens.

        Returns:
            bool: True is the authentication object contains valid information, else False.
        """
        # verify if tokens are valid
        if time.gmtime() > time.struct_time(auth_object.refresh_expires_time):
            return False
        # verify if the domain of cached tokens corresponds to the desire domain
        if self.domain != auth_object.domain:
            return False
        # verify if the scope of cached tokens corresponds to the desire scope
        if scope is not None and scope not in auth_object.scope:
            return False
        return True

    def _get_elg_execution_location(self):
        """
        Method to get the elg execution location information from the metadata record.
        """
        self.elg_execution_location_sync = get_information(
            id=self.id, obj=self.record, infos=["service_info", "elg_execution_location_sync"]
        )
        self.elg_execution_location = get_information(
            id=self.id, obj=self.record, infos=["service_info", "elg_execution_location"]
        )

    def _get_resource_info(self):
        """
        Method to get the resource information from the metadata record.

        Raises:
            ValueError: The content type in the input media type is not detected.
        """
        # input
        input_content_resource = get_information(
            id=self.id, obj=self.record, infos=["described_entity", "lr_subclass", "input_content_resource"]
        )
        assert len(input_content_resource) == 1, f"Entity {self.id} has multiple input content resource"
        input_content_resource = input_content_resource[0]
        self.processing_resource_type = input_content_resource.get("processing_resource_type")
        self.input_media_type = input_content_resource.get("media_type")
        self.input_data_format = input_content_resource.get("data_format")
        self.input_character_encoding = input_content_resource.get("character_encoding")
        if "text" in self.input_media_type:
            self.default_content_type = "text/plain"
        elif "audio" in self.input_media_type:
            self.default_content_type = "audio/x-wav"
        else:
            raise ValueError(f"Content-Type not detected in {self.input_media_type}")
        self._get_opening_mode()
        # output
        output_content_resource = get_information(
            id=id, obj=self.record, infos=["described_entity", "lr_subclass", "output_resource"]
        )
        assert len(output_content_resource) == 1, f"Entity {self.id} has multiple output content resource"
        output_content_resource = output_content_resource[0]
        self.output_media_type = output_content_resource.get("media_type")
        self.output_data_format = output_content_resource.get("data_format")
        self.output_character_encoding = output_content_resource.get("character_encoding")

    def _get_content_type(self, filename: str) -> str:
        """
        Method to get the content type of a file.

        Args:
            filename (str): name of the file.

        Returns:
            str: content type of the file
        """
        _, extension = splitext(filename)
        content_type = MIME.get_content_type(extension)
        if content_type is None:
            content_type = self.default_content_type
        return content_type

    def _get_opening_mode(self):
        """
        Method to get the opening mode of the service.

        Raises:
            ValueError: the character encoding is not recognized.
        """
        if self.input_character_encoding is None or (
            isinstance(self.input_character_encoding, list) and len(self.input_character_encoding) == 0
        ):
            mode = "rb"
        elif isinstance(self.input_character_encoding, list):
            assert len(self.input_character_encoding) == 1, f"Entity {self.id} has multiple input character encoding"
            if "UTF-8" in self.input_character_encoding[0]:
                mode = "r"
            else:
                raise ValueError(f"Character encoding ({self.input_character_encoding}) not recognized.")
        else:
            raise ValueError(f"Character encoding ({self.input_character_encoding}) not recognized.")
        self.opening_mode = mode

    @need_authentication()
    def __call__(
        self,
        input_str: str = None,
        sync_mode: bool = False,
        timeout: int = None,
        check_file: bool = True,
        verbose: bool = True,
        output_func: Union[str, Callable] = lambda response: response,
    ) -> Union[dict, str]:
        """
        Method to call a service. You can enter a string input or the path to the file to process.
        The output is returned in JSON format.

        ```
        response = service("This is a string text to process")
        response = service("/path/to/the/file/to/process")
        ```

        Args:
            input_str (str, optional): can be the text to process directly or the name of the file to process.
            Defaults to None.
            sync_mode (bool, optional): True to use the sync_mode. Defaults to False.
            timeout (int, optional): number of seconds before timeout. Defaults to None.
            check_file (bool, optional): True to check if input_str can be a file or not. Defaults to True.
            verbose (bool, optional): False to avoid print messages. Defaults to True.
            output_func (Union[str, Callable], optional): function applied to the service response. It can be used
                to extract only the content from the response. If set to 'auto', a generic extractive function will
                be used. Defaults to lambda response: response.

        Raises:
            ValueError: the input_str is None.
            ElgException: can raise a specific Elg exception if the request to the service did not succeed.

        Returns:
            Union[dict, str]: service response in JSON format or as a string if output_func returns a string.
        """
        if input_str is None:
            raise ValueError("You need to provide at least a filename or text as string.")

        if isinstance(output_func, str) and output_func == "auto":
            output_func = get_content_from_response
        if not callable(output_func):
            raise ValueError("output_func should be a callable object or 'auto'.")

        if verbose:
            print(f"Calling:\n\t[{self.id}] {self.resource_name}")

        headers = {
            "Authorization": f"Bearer {self.authentication.access_token}",
        }
        if check_file and isfile(input_str):
            headers["Content-Type"] = self._get_content_type(input_str)
            if verbose:
                print(f"Found filename as input:\n\t{input_str}\n")
            with open(input_str, self.opening_mode) as f:
                if sync_mode:
                    return self._call_sync(f=f, headers=headers, timeout=timeout, output_func=output_func)
                else:
                    return self._call_async(f=f, headers=headers, timeout=timeout, output_func=output_func)
        else:
            headers["Content-Type"] = "text/plain"
            if verbose:
                print(f"Found plain text as input:\n\t{input_str}\n")
            with io.StringIO(input_str) as f:
                if sync_mode:
                    return self._call_sync(f=f, headers=headers, timeout=timeout, output_func=output_func)
                else:
                    return self._call_async(f=f, headers=headers, timeout=timeout, output_func=output_func)

    @catch_requests_error
    def _call_sync(self, f: io.IOBase, headers: dict, timeout: int, output_func: Callable) -> Union[dict, str]:
        """
        Method used by __call__ to call a service synchronously.

        Args:
            f (io.IOBase): input file.
            headers (dict): headers for the call.
            timeout (int): number of seconds before timeout.
            output_func (Union[str, Callable], optional): function applied to the service response. It can be used
                to extract only the content from the response. If set to 'auto', a generic extractive function will
                be used. Defaults to lambda response: response.

        Returns:
            Union[dict, str]: service response in JSON format or as a string if output_func returns a string.
        """
        response = requests.post(self.elg_execution_location_sync, headers=headers, data=f, timeout=timeout)
        ensure_response_ok(response)
        return output_func(response.json())

    @catch_requests_error
    def _call_async(self, f: io.IOBase, headers: dict, timeout: int, output_func: Callable) -> Union[dict, str]:
        """
        Method used by __call__ to call a service asynchronously.

        Args:
            f (io.IOBase): input file.
            headers (dict): headers for the call.
            timeout (int): number of seconds before timeout.
            output_func (Union[str, Callable], optional): function applied to the service response. It can be used
                to extract only the content from the response. If set to 'auto', a generic extractive function will
                be used. Defaults to lambda response: response.

        Returns:
            Union[dict, str]: service response in JSON format or as a string if output_func returns a string.
        """
        response = requests.post(self.elg_execution_location, headers=headers, data=f, timeout=timeout)
        ensure_response_ok(response)
        response = response.json()["response"]
        assert response["type"] == "stored"
        headers.pop("Content-Type")
        uri = response["uri"]
        response = requests.get(uri, headers=headers)
        waiting_time = time.time()
        while response.ok and "progress" in response.json().keys():
            sleep(1)
            response = requests.get(uri, headers=headers, timeout=timeout)
            if time.time() - waiting_time > (timeout if timeout is not None else float("inf")):
                raise requests.exceptions.Timeout("The service didn't respond before the end of the timeout.")
        ensure_response_ok(response)
        return output_func(response.json())
